#include <string.h>
#include <stdio.h>
#include "Map.h"
#include "Robot.h"

extern MapClass MyMap;
extern RobotGangClass * MyRobot;
extern PackageClass PackageList;

static char Buffer[10000];

static char LastCommand[100];

static MoveType LastMove = None;

float Dist (int x1, int y1, int x2, int y2) {
  return (float)sqrt(abs((x1*x1)-(x2*x2)) + abs((y1*y1)-(y2*y2)));
}

PackageClass * DeletePackage(PackageClass * PP) {
  if (PackageList.FirstPackage == PP){
    PackageList.FirstPackage = PP->NextPackage;
    delete PP;
    return PackageList.FirstPackage;
  }
  PackageClass * TempPP = PackageList.FirstPackage;
  while (TempPP->NextPackage != PP)
    TempPP = TempPP->NextPackage;
  TempPP->NextPackage = PP->NextPackage;
  delete PP;
  return TempPP->NextPackage;
}

void SendCommand(FILE * SockStream, char *Command) {

  strcpy (LastCommand, Command);
  fprintf(SockStream, "%s\n", Command);
}

PackageClass * PackageClass::FirstPackage = NULL;

PackageClass * PackageClass::AddPackage(int IDp, int Xp, int Yp,
					int Dx, int Dy, int W) {
  PackageClass * PP = new PackageClass;
  PP->NextPackage = FirstPackage;
  PP->ID = IDp;
  PP->Weight = W;
  PP->DestX = Dx;
  PP->DestY = Dy;
  PP->XPos = Xp;
  PP->YPos = Yp;
  PP->InTransit = false;
  PP->Value = Weight / sqrt(abs(Dx-Xp)+abs(Dy-Yp)); // should be Astar
  FirstPackage = PP;
  PP->DeleteMe = false;
  return PP;
}

PackageClass * PackageClass::FindPackage(int IDp) {
  PackageClass * PP = FirstPackage;
  while (PP != NULL) {
    if (PP->ID == IDp)
      return PP;
    PP = PP->NextPackage;
  }
  return NULL;
}

PackageClass * PackageClass::GetNextPackage(void) {
  PackageClass * BestPackage = NULL;
  PackageClass * PP = FirstPackage;
  
  while (PP != NULL) {
    if ((PP->XPos == MyRobot->Robot->GetXPos())&&
	(PP->YPos == MyRobot->Robot->GetYPos())&&
	(PP->InTransit == false) &&
	(PP->Weight < (MyRobot->Robot->MaxWeight - MyRobot->Robot->CurrentWeight))) {
      if (BestPackage == NULL) 
	BestPackage = PP;
      else if (BestPackage->Value < PP->Value)
	BestPackage = PP;
    }
    PP=PP->NextPackage;
  }
  if (BestPackage != NULL)
    BestPackage->InTransit = true;
  return BestPackage;
}

PackageClass * PackageClass::GetNextPackageTo(int X, int Y) {
  PackageClass * BestPackage = NULL;
  PackageClass * PP = PackageList.FirstPackage;
  
  while (PP != NULL) {
    if ((PP->XPos == MyRobot->Robot->GetXPos())&&
	(PP->YPos == MyRobot->Robot->GetYPos())&&
	(PP->InTransit == false) &&
	(PP->Weight < (MyRobot->Robot->MaxWeight - MyRobot->Robot->CurrentWeight)) &&
	(PP->DestX == X) &&
	(PP->DestY == Y)) {
      if (BestPackage == NULL) 
	BestPackage = PP;
      else if (BestPackage->Value < PP->Value)
	BestPackage = PP;
    }
    PP=PP->NextPackage;
  }
  if (BestPackage != NULL)
    BestPackage->InTransit = true;
  return BestPackage;
}
  
    
void PackageClass::ProcessPackageList(char * List) {

  PackageClass * PP;

  // set all packages at this location for deletion
  PP = PackageList.FirstPackage;
  while (PP != NULL) {
    if ((PP->XPos == MyRobot->Robot->XPos) &&
	(PP->YPos == MyRobot->Robot->YPos) &&
	(PP->InTransit == false))
      PP->DeleteMe = true;
    PP = PP->NextPackage;
  }

  printf("Finished marking, processing list   \n");
  //printf("Package list is |%s|\n", List);
  if ((List == NULL) || (strlen(List)<2))
    return;
  char * Element;
  Element = strtok(List, " ");
  int ID, DestX, DestY, Weight;

   while (Element != NULL) {
     sscanf(Element, "%i", &ID);
     Element = strtok(NULL, " ");
     if (Element == NULL) break;
     sscanf(Element, "%i", &DestX);
     Element = strtok(NULL, " ");
     if (Element == NULL) break;
     sscanf(Element, "%i", &DestY);
     Element = strtok(NULL, " ");
     if (Element == NULL) break;
     sscanf(Element, "%i", &Weight);
     PP = FindPackage(ID);
     if (PP == NULL)
       PP = AddPackage(ID,
		  MyRobot->Robot->GetXPos(),
		  MyRobot->Robot->GetYPos(),
		  DestX, DestY, Weight);
     else {
       // refresh package data in case it changed
       PP->XPos = MyRobot->Robot->GetXPos();
       PP->YPos = MyRobot->Robot->GetYPos();
       PP->DestX = DestX;
       PP->DestY = DestY;
       PP->Weight = Weight;
       PP->InTransit = false;
       PP->DeleteMe = false;
     }
       
     Element = strtok(NULL, " ");
   }  

   if (PackageList.FirstPackage == NULL)
     return;

   printf("Cleaning up missing packages    \n");
   // delete at head
   while (PackageList.FirstPackage->DeleteMe) {
     PP = PackageList.FirstPackage;
     PackageList.FirstPackage = PP->NextPackage;
     delete PP;
     if (PackageList.FirstPackage == NULL)
       return;
   }
   
   PackageClass * PPP = PackageList.FirstPackage;
   PP = PPP->NextPackage;
   while (PP != NULL) {
     if (PP->DeleteMe) {
       PPP->NextPackage = PP->NextPackage;
       delete PP;
       PP = PPP->NextPackage;
     }
     else {
       PPP=PP;
       PP = PP->NextPackage;
     }
   }
}



void RobotGangClass::ProcessStatus(FILE * SockStream) {

  fgets(Buffer, 10000, SockStream);
  //printf("Status line: |%s|\n", Buffer);
  
  if (Buffer[0] != '#') {
    // looks like game is over, print out final messages and exit
    printf(Buffer);
    int linecount = 0;
    while (1) {
      fgets(Buffer, 10000, SockStream);
      if ((strlen(Buffer) <2) || (feof(SockStream)) || (linecount++ > 7))
	exit(0);
      printf(Buffer);
    }
  }

  // mark all robots for deletion
  RobotGangClass* RP = FirstRobot;
  while (RP != NULL) {
    RP->DeleteMe = true;
    RP = RP->NextRobot;
  }

  RobotCount = 0;
  ProcessRobot(Buffer);


  // delete all robots not updated
  while (FirstRobot->DeleteMe) {
    RP = FirstRobot;
    FirstRobot = RP->NextRobot;
    delete RP->Robot;
    delete RP;
  }
  
  RobotGangClass * PRP = FirstRobot;
  RP = PRP->NextRobot;
  while (RP != NULL) {
    if (RP->DeleteMe) {
      PRP->NextRobot = RP->NextRobot;
      delete RP->Robot;
      delete RP;
      RP = PRP->NextRobot;
    }
    else {
      PRP = RP;
      RP = RP->NextRobot;
    }
  }
}

void RobotGangClass::ProcessRobot(char * Line) {

  printf("Processing Robot commands\n");
  RobotGangClass * RP;
  PackageClass * PP;

  char * Cmd = strtok(Line, " \n");
  int Value;
//   int i;

  while (Cmd != NULL) {
    switch (*Cmd) {
    case 'N':
      RP->Robot->ProcessMovement(RobotClass::North);
      break;
    case 'S':
      RP->Robot->ProcessMovement(RobotClass::South);
      break;
    case 'E':
      RP->Robot->ProcessMovement(RobotClass::East);
      break;
    case 'W':
      RP->Robot->ProcessMovement(RobotClass::West);
      break;
    case 'X':
      sscanf(strtok(NULL, " "), "%i", &Value);
      RP->Robot->SetXPos(Value);
      break;
    case 'Y':
      sscanf(strtok(NULL, " "), "%i", &Value);
      RP->Robot->SetYPos(Value);
      break;
    case 'P':
      // Mark Packages in transit when picked up
      sscanf(strtok(NULL, " "), "%i", &Value);
      PP = PackageList.FindPackage(Value);
      if (PP != NULL)
	PP->InTransit = true;
      break;
    case 'D':
      // if we are processing our own response
      // and we dropped a package, update our 
      // carried package list
//       if (RP->Robot->PlayerID == MyRobot->Robot->PlayerID) {
// 	sscanf(strtok(NULL, " "), "%i", &Value);
// 	for (i=0; i<20; i++) {
// 	  if ((MyRobot->Robot->Package[i] != NULL) &&
// 	      (MyRobot->Robot->Package[i]->ID == Value)) {
// 	    MyRobot->Robot->Package[i] = NULL;
// 	    MyRobot->Robot->CompressPackageList();
// 	  }
// 	}
//       }
      break;
    case '#':
      RobotCount++;
      // another robot 
      // get the robot, if it doesn't exist, create it
      sscanf(Cmd+1, "%i", &Value);
      //printf("Processing robot %i\n", Value);
      RP = FindRobot(Value);
      if (RP == NULL) {
	//printf("robot not found, adding robot \n");
	RP = AddRobot(Value, &MyMap);
      }
      RP -> DeleteMe = false;
      break;
      
    default:
      break;
    }
    Cmd = strtok(NULL, " ");
  }
}


RobotGangClass * RobotGangClass::FirstRobot = NULL;
int RobotGangClass::RobotCount = 0;

RobotGangClass * RobotGangClass::RobotAt(int Xp, int Yp) {
  RobotGangClass * RP = FirstRobot;
  while (RP != NULL) {
    if ((RP->Robot->XPos == Xp) &&
	(RP->Robot->YPos == Yp))
      return RP;
    RP = RP->NextRobot;
  }
  return NULL;
}


RobotGangClass * RobotGangClass::FindRobot(int ID){
  RobotGangClass * RP = FirstRobot;
  while (RP != NULL){
    if (RP->Robot->GetPlayerID() == ID)
      return RP;
    RP = RP->NextRobot;
  }
  return NULL;
}

RobotGangClass * RobotGangClass::AddRobot(int ID, MapClass * Map){
  RobotGangClass * RP = new RobotGangClass;
  RP->Robot = new RobotClass;
  RP->Robot->SetMap(Map);
  RP->Robot->SetPlayerID(ID);
  RP->Updated = false;
  RP->DeleteMe = false;
  RP->NextRobot = FirstRobot;
  FirstRobot = RP;
  int i;
  for (i=0; i<20; i++)
    RP->Robot->Package[i] = NULL;
  RP->Robot->GoalX = 0;
  RP->Robot->GoalY = 0;
  RP->Robot->Mission = DeliveringPackages;
  return RP;
}

void RobotGangClass::CleanList(void) {
  printf ("I shouldn't be called\n");
  exit(1);
//   RobotGangClass * RP = FirstRobot;
//   while ((RP->Updated == false) && (RP != NULL)){
//     delete FirstRobot->Robot;
//     delete FirstRobot;
//     FirstRobot = RP->NextRobot;
//     RP = FirstRobot;
//   }

//   if (RP!= NULL)
//     RP = RP->NextRobot;
//   RobotGangClass * PRP = FirstRobot;
//   while (RP != NULL) {
//     if (RP->Updated==false) {
//       // robot marked for elimination :)
//       PRP->NextRobot = RP->NextRobot;
//       delete RP->Robot;
//       delete RP;
//     }
//     else {
//       PRP = PRP->NextRobot;
//     }
//     RP = RP->NextRobot;
//   } 
}

void RobotGangClass::ResetUpdated(void) {
  printf ("I shouldn't be called\n");
  exit(1);
//   RobotGangClass * RP = FirstRobot;
//   while (RP != NULL) {
//     RP->Updated = false;
//     RP = RP->NextRobot;
//   }
}
	


void RobotClass::ProcessPackages(FILE * SockStream) {
  fgets(Buffer, 10000, SockStream);
  PackageList.ProcessPackageList(Buffer);
}

void RobotClass::CompressPackageList(void) {
  int OpenPointer, NextPointer ;
  OpenPointer = 0; NextPointer = 1;
  while (1) {
    if (Package[OpenPointer] == NULL) {
      while (Package[NextPointer] == NULL)
	if (++NextPointer >= 20){
	  return;
	}
      Package[OpenPointer++] = Package[NextPointer];
      Package[NextPointer++] = NULL;
    } 
    else {
      OpenPointer++;
      NextPointer++;
    }
  }
}

void RobotClass::MoveRandom(FILE * SockStream, int Bid) {
  
  //printf("Moving random direction\n");
  int trycount = 0;
  char Command[20];
  while (trycount++ < 20) {
    switch (rand() % 4) {
    case 0:
      // Move North
      if (MyMap.FreeTile(XPos, YPos+1)) {
	sprintf(Command, "%i Move N", Bid);
	SendCommand(SockStream, Command);
	Energy -= Bid;
	return;
      }
    case 1:
      // Move South
      if (MyMap.FreeTile(XPos, YPos-1)) {
	sprintf(Command, "%i Move S", Bid);
	SendCommand(SockStream, Command);
	Energy -=Bid;
	return;
      }
    case 2:
      // Move East
      if (MyMap.FreeTile(XPos+1, YPos)) {
	sprintf(Command, "%i Move E", Bid);
	SendCommand(SockStream, Command);
	Energy -=Bid;
	return;
      }
    case 3:
      // Move West
      if (MyMap.FreeTile(XPos-1, YPos)) {
	sprintf(Command, "%i Move W", Bid);
	SendCommand(SockStream, Command);
	Energy -=Bid;
	return;
      }
    }
  }
  printf("*****  HELP, I've Fallen and I cant get up *****\n");
  SendCommand(SockStream, "1 Pick 99");
  Energy -= 1;
  return;
}      


void RobotClass::GenerateCommand(FILE * SockStream) {
  DirectionType Dir;
  float Cost;
  char Command[200];
  Command[0]=0;

  

  // AI loop
  // Do we need to do special moves?
  //   Escape move
  //   Deathbot move
  //   Dumbass move
  // Are we standing on any good packages?
  //   pick up the packages
  // Do we have any packages to deliver?
  //   Deliver best value package
  //   Remove that package from the package list
  // Are there any packages in the package list?
  //   Move to best value package
  // Move to a random base


  // escape sequence move
  // if we are between a robot and water
  // we need to spend some $$ to get out
  // of the way.  Since we would be out of
  // the game and lose our $$ we should spend
  // an amount relative to our $$
  // don't do more than 2 consecutive escapes, 
  // Do a move random after that.

  if (MyRobot->RobotCount > 1) {
    static int escapecount = 0;
    // Water N
    int Bid = Energy/30; // (3.33 % of total energy)
    if ((MyMap.TileType(XPos, YPos+1) == 2) &&
	(MyRobot->RobotAt(XPos, YPos-1))) {
      if (escapecount > 1) {
	MoveRandom(SockStream, Bid);
      }
      else {
	sprintf(Command, "%i Move S", Bid);
	SendCommand(SockStream, Command);
	Energy -= Bid;
	LastMove = Escape;
	escapecount++;
	return;
      }
    }
    
    // Water S
    Bid = Energy/30; // (3.33 % of total energy)
    if ((MyMap.TileType(XPos, YPos-1) == 2) &&
	(MyRobot->RobotAt(XPos, YPos+1))) {
      if (escapecount > 1) {
	MoveRandom(SockStream, Bid);
      }
      else {
	sprintf(Command, "%i Move N", Bid);
	SendCommand(SockStream, Command);
	Energy -= Bid;
	LastMove = Escape;
	escapecount++;
	return;
      }
    }
	
    // Water E
    Bid = Energy/30; // (3.33 % of total energy)
    if ((MyMap.TileType(XPos+1, YPos) == 2) &&
	(MyRobot->RobotAt(XPos-1, YPos))) {
      if (escapecount > 1) {
	MoveRandom(SockStream, Bid);
      }
      else {
	sprintf(Command, "%i Move W", Bid);
	SendCommand(SockStream, Command);
	Energy -= Bid;
	LastMove = Escape;
	escapecount++;
	return;
      }
    }
	
    // Water W
    Bid = Energy/30; // (3.33 % of total energy)
    if ((MyMap.TileType(XPos-1, YPos) == 2) &&
	(MyRobot->RobotAt(XPos+1, YPos))) {
      if (escapecount > 1) {
	MoveRandom(SockStream, Bid);
      }
      else {
	sprintf(Command, "%i Move E", Bid);
	SendCommand(SockStream, Command);
	Energy -= Bid;
	LastMove = Escape;
	escapecount++;
	return;
      }
    }

    escapecount = 0;
  }
  
  // deathbot move
  // if a robot is between us and water, 
  // let's try to push him in.

  const int Push = 20;

  
  if (MyRobot->RobotCount > 1) {
    // to the north
    if ((MyMap.TileType(XPos, YPos +2) == 2) &&
	(MyRobot->RobotAt(XPos, YPos +1)) &&
	(MyMap.TileType(XPos, YPos +1) == 0)) {
      sprintf(Command, "%i Move N", Push);
      SendCommand(SockStream, Command);
      Energy -= Push;
      LastMove = Deathbot;
      return;
    }
    
    // to the South
    if ((MyMap.TileType(XPos, YPos -2) == 2) &&
	(MyRobot->RobotAt(XPos, YPos -1)) &&
	(MyMap.TileType(XPos, YPos -1) == 0)) {
      sprintf(Command, "%i Move S", Push);
      SendCommand(SockStream, Command);
      Energy -= Push;
      LastMove = Deathbot;
      return;
    }
    
    // to the east
    if ((MyMap.TileType(XPos +2, YPos) == 2) &&
	(MyRobot->RobotAt(XPos+1, YPos)) &&
	(MyMap.TileType(XPos+1, YPos) == 0)) {
      sprintf(Command, "%i Move E", Push);
      SendCommand(SockStream, Command);
      Energy -= Push;
      LastMove = Deathbot;
      return;
    }
    
    // to the west
    if ((MyMap.TileType(XPos-2, YPos) == 2) &&
	(MyRobot->RobotAt(XPos-1, YPos)) &&
	(MyMap.TileType(XPos-1, YPos) == 0)) {
      sprintf(Command, "%i Move W", Push);
      SendCommand(SockStream, Command);
      Energy -= Push;
      LastMove = Deathbot;
      return;
    }
  }      
  
  
  
  // get around dumbass robots :
  //   if we're stuck in the same place for more than
  //   5 moves, lets try to move in a random direction
  static int PrevX = 0;
  static int PrevY = 0;
  static int Count = 0;
  
  if ((abs(PrevX-XPos) >2) ||
      (abs(PrevY-YPos) >2) ||
      (MyRobot->RobotCount == 1)) {
    PrevX = XPos;
    PrevY = YPos;
    Count = 0;
  }
  else if (Count++ > 8) {
    MoveRandom(SockStream);
    LastMove = Dumbass;
    return;
  }
  
  int NextPackIndex=0;
  while (Package[NextPackIndex++] != NULL);
  
  // compress carried package list
  //printf("compressing package list\n");
  CompressPackageList();
  
  if (Package[0] == NULL) {
    // clean up world Package list since we are not carrying any
    PackageClass * PP = PackageList.FirstPackage;
    while (PP != 0) {
      if (PP->InTransit) {
	PP=DeletePackage(PP);
      }
      else {
	// recompute the value of the package -- Just a rough estimate,
	// the directed distance to the package + distance of package to
	// the package destination
	PP->Value = PP->Weight / (Dist(XPos, YPos, PP->XPos, PP->YPos) +
				  Dist(PP->XPos, PP->YPos, PP->DestX, PP->DestY));
	PP= PP->NextPackage;
      }
    }
  }

  // update mission
  if ((Mission == DeliveringPackages) && (Package[0] ==NULL)){
    // out of packages, we need to find some packages
    // either go to packages in our list, or go to a random base
    printf("Finding new mission \n");
    GoalX = 0; 
    Mission = MovingToBase;
    while (GoalX == 0){
      int trycount = 0;
      if ((PackageList.FirstPackage == NULL) || ((rand()% 5==0))) {
	// move to random base
	printf("Finding random base     \n");
	while (((XPos == GoalX) && (YPos == GoalY)) || (GoalX == 0)) {
	  MyMap.FindBase(&GoalX, &GoalY);
	  if(trycount++ > 20) {
	    break;
	  }
	}
	if (trycount > 20) {
	  printf("random base not found, moving at random  \n");
	  Mission = MovingRandom;
	  MoveRandom(SockStream);
	  exit(0);
	  return;
	}
      }
      else {
	// finding best package
	printf("Finding Package from world Package list    \n");
	PackageClass * Best = PackageList.FirstPackage;
	PackageClass * PP = PackageList.FirstPackage->NextPackage;
	while (PP != NULL) {
	  if (PP->Value > Best->Value)
	    Best = PP;
	  PP = PP->NextPackage;
	}
	GoalX = Best->XPos;
	GoalY = Best->YPos;
      }
    }
    if (GoalX == 0) {
      // no place to go
      // Move in random direction
      printf("Could not find new misson, moving at random  \n");
      Mission = MovingRandom;
      MoveRandom(SockStream);
    }
  }

  if (Package[0] == NULL) {
    //printf("getting best package\n ");
    // get best package at this location
    Package[0] = PackageList.GetNextPackage();
    if (Package[0] != NULL) {
      sprintf(Command, "1 Pick %i ", Package[0]->ID);
      CurrentWeight += Package[0]->Weight;
      NextPackIndex++;
    }
  }

  // get other packages to primary destination
  while ((NextPackIndex < 20) && 
	 (Package[NextPackIndex] == NULL) &&
	 (Package[0] != NULL)) {
    //printf("getting other packages at same dest\n");
    Package[NextPackIndex] = PackageList.GetNextPackageTo(Package[0]->DestX,
							  Package[0]->DestY);
    if (Package[NextPackIndex] == NULL)
      break;
    CurrentWeight += Package[NextPackIndex]->Weight;
    if (Command[0] == 0)
      sprintf(Command, "1 Pick %i ", Package[NextPackIndex]->ID);
    else{
      char TempStr[10];
      sprintf(TempStr, "%i ", Package[NextPackIndex]->ID);
      strcat(Command, TempStr);
    }
    NextPackIndex++;
  }

  // get other packages to anywhere
  while ((NextPackIndex < 20) &&
	 (Package[NextPackIndex] ==NULL)) {
    //printf("getting other packages at different dest\n");
    Package[NextPackIndex] = PackageList.GetNextPackage();
    if (Package[NextPackIndex] == NULL)
      break;
    CurrentWeight += Package[NextPackIndex]->Weight;
    if (Command[0] == 0)
      sprintf(Command, "1 Pick %i ", Package[NextPackIndex]->ID);
    else{
      char TempStr[10];
      sprintf(TempStr, "%i ", Package[NextPackIndex]->ID);
      strcat(Command, TempStr);
    }
    NextPackIndex++;
  }
    
  // issue pickup command
  if (Command[0] != 0) {
    SendCommand(SockStream, Command);
    Energy--;
    Mission = DeliveringPackages;
    return;
  }

  // do we need to drop any packages?
  for (NextPackIndex=0; NextPackIndex<20; NextPackIndex++) {
    if ((Package[NextPackIndex] != NULL) &&
	(Package[NextPackIndex]->DestX == XPos) &&
	(Package[NextPackIndex]->DestY == YPos)) {
      if (Command[0] == 0)
	sprintf(Command, "1 Drop %i ", Package[NextPackIndex]->ID);
      else {
	char TempStr[10];
	sprintf(TempStr, "%i ", Package[NextPackIndex]->ID);
	strcat(Command, TempStr);
      }
      CurrentWeight -= Package[NextPackIndex]->Weight;
      Package[NextPackIndex] = NULL;
    }
  }

  // issue drop command
  if (Command[0] != 0) {
    //printf("dropping packages\n");
    SendCommand(SockStream, Command);
    Energy--;
    return;
  }

  //printf("Compressing package list\n");
  CompressPackageList();

  // go deliver package
  if (Package[0] != NULL) {
    //printf("Moving to package delivery location\n");
    GoalX = Package[0]->DestX;
    GoalY = Package[0]->DestY;
    Mission = DeliveringPackages;
  }

  // if we reach the base and there are not packages
  // we need to go to another base
  if ((Mission == MovingToBase) &&
      (GoalX == XPos) &&
      (GoalY == YPos)) {
    int basecount = 0;
    while (((XPos == GoalX) && (YPos == GoalY)) || (GoalX == 0)) {
      MyMap.FindBase(&GoalX, &GoalY);
      if(basecount++ > 30) {
	break;
      }
    }
    if  (basecount > 30) {
      GoalX = 0;
    }
  }

  if (GoalX != 0) {
    //printf("computing route to goal (%i, %i), (%i, %i)\n", XPos, YPos, GoalX, GoalY);
    Dir = AStarNextMove(XPos, YPos, GoalX, GoalY, &Cost);
    //printf("Goalseek done\n");

    // If a robot is in our destination, bid 2, otherwise bid 1
    switch (Dir) {
    case North:
      if (MyRobot->RobotAt(XPos, YPos+1)) 
	SendCommand(SockStream, "2 Move N");
      else
	SendCommand(SockStream, "1 Move N");
      break;
    case South:
      if (MyRobot->RobotAt(XPos, YPos-1)) 
	SendCommand(SockStream, "2 Move S");
      else
	SendCommand(SockStream, "1 Move S");
      break;
    case East:
      if (MyRobot->RobotAt(XPos+1, YPos)) 
	SendCommand(SockStream, "2 Move E");
      else
	SendCommand(SockStream, "1 Move E");
      break;
    case West:
      if (MyRobot->RobotAt(XPos-1, YPos)) 
	SendCommand(SockStream, "2 Move W");
      else
	SendCommand(SockStream, "1 Move W");
      break;
    case Stay:
      SendCommand(SockStream, "1 Pick 99");
      break;
    }
    Energy--;
  }
  else{
    SendCommand(SockStream, "1 Pick 123");
  }
}


void RobotClass::ProcessMovement(DirectionType Direction) {
  switch (Direction) {
  case North: 
    YPos++;
    break;
  case South :
    YPos--;
    break;
  case East:
    XPos++;
    break;
  case West:
    XPos--;
    break;
  }
};

void RobotClass::DetermineAction(void) {
  printf("I should not be called\n");
  exit(1);
//   DirectionType Dir;
//   float Cost;
//   while ((XPos != GoalX) || (YPos != GoalY)) {
//     Dir = AStarNextMove(XPos, YPos, GoalX, GoalY, &Cost);
//     ProcessMovement(Dir);
//     Print();
//   }
};

void RobotClass::Print(void) {
  int i, j;
  static int count = 0;
  RobotGangClass * RP;
  printf("\x1b[H");
  printf("Mission : ");
  switch(MyRobot->Robot->Mission) {
  case MovingToBase :
    printf("Moving to Base       ");
    break;
  case DeliveringPackages :
    printf("Delivering Packages  ");
    break;
  case  MovingRandom :
    printf("Moving at Random     ");
    break;
  default:
    printf("***Error***          ");
  }
  printf("Package[0]  0x%08x  ", (unsigned) MyRobot->Robot->Package[0]); 
  if (MyRobot->Robot->Package[0] ==NULL) {
    printf("(xxx,xxx)  ");
  }
  else{
    printf("(%3i,%3i)  ", MyRobot->Robot->Package[0]->DestX, 
	   MyRobot->Robot->Package[0]->DestY);
  }
    
  printf("\n");
  printf("Cur Goal: (%3i, %3i)  ", MyRobot->Robot->GoalX, MyRobot->Robot->GoalY);
  printf("         First Package: 0x%08x  ", (unsigned) PackageList.FirstPackage);
  if (PackageList.FirstPackage ==NULL) {
    printf("Dest:(xxx,xxx)  ");
    printf("Loc: (xxx,xxx)  ");
  }
  else{
    printf("Dest:(%3i,%3i)  ", PackageList.FirstPackage->DestX, 
	   PackageList.FirstPackage->DestY);
    printf("Loc: (%3i,%3i)  ", PackageList.FirstPackage->XPos, 
	   PackageList.FirstPackage->YPos);
  }
    
  printf("\n");
  printf("Engergy: %i          ",MyRobot->Robot->Energy);
  printf("Special Move: ");
  switch (LastMove) {
  case Deathbot :
    printf("Deathbot      ");
    break;
  case Escape:
    printf("Escape        ");
    break;
  case Dumbass:
    printf("Evade         ");
    break;
  case None:
    printf("None          ");
    break;
  default:
    printf("**invalid***  ");
  }

  printf("Robot ID:%4i  ", MyRobot->Robot->PlayerID);
  printf("Robot Count:%4i  ", MyRobot->RobotCount);

  printf("\n");
  printf("Last Command: %-80s\n", LastCommand);
  printf("      +"); 
  for(i=Map->GetMapSizeX(); i>=1; i--) 
    printf("-"); 
  printf("+\n");
  for (i=Map->GetMapSizeY(); i>=1; i--) {
    printf ("R %3i |", i);
    for (j=1; j<= Map->GetMapSizeX(); j++) {
      if ((j == XPos) && (i == YPos))
	printf("R");
      else if ((j == GoalX) && (i == GoalY))
	printf("G");
      else if (MyRobot->RobotAt(j, i)){
	RP = MyRobot->RobotAt(j, i);
	int TempID = RP->Robot->PlayerID;
	char TempString[5];
	sprintf(TempString, "%i", TempID % 10);
	printf(TempString);
      }
      else {
	int Tile = Map->TileType(j, i);
	switch (Tile) {
	case 0: 
	  printf(".");
	  break;
	case 1:
	  printf("#");
	  break;
	case 2 :
	  printf("~");
	  break;
	case 3:
	  printf("@");
	  break;
	default:
	  printf("?");
	}
      }
    }
    printf ("|\n");
  }
  printf("      +"); 
  for(i=1; i<= Map->GetMapSizeX(); i++) 
    printf("-"); 
  printf("+ %i \n", count++);
}

RobotClass::DirectionType RobotClass::AStarNextMove(int x1, int y1, int x2, int y2, float * Cost) { 

  NODE *path, *Node; 
  DirectionType RetVal;

  // find the path
  if ((x1 == x2) && (y1 == y2))
    return Stay;
  path=(NODE *)FindPath(x1, y1, x2, y2); 
  
  if (path == NULL) {
    printf ("***** no path from (%i, %i) to (%i, %i)      \n", x1, y1, x2, y2);
    exit(1);
  }

  *Cost = path->g;
  // path is destination -> origin we need to find the next to last node
  while (path -> Parent -> Parent !=NULL) {
    path = path -> Parent;
  }

  // Go East
  if (path->x == XPos + 1)
    RetVal = East;
  // Go West
  if (path->x == XPos -1)
    RetVal = West;
  // Go North
  if (path->y == YPos +1)
    RetVal = North;
  // Go South
  if (path->y == YPos -1)
    RetVal = South;

  // Clean up memory
  Node=OPEN->NextNode; 
  while (Node != NULL) { 
    free(Node);
    Node=Node->NextNode; 
  } 

  Node=CLOSED->NextNode; 
  while (Node != NULL) {
    free(Node); 
    Node=Node->NextNode; 
  } 

  return RetVal;
}

/////////////////////////////////////////////////////////////////////////
// Algorithm to find shortest path (A*)
/////////////////////////////////////////////////////////////////////////

// A* Algorithm struct NODE*
NODE * RobotClass::FindPath(int sx, int sy, int dx, int dy) {
  lPropagateDown=lPropLoop=lGenerateSuccessors=
    lGenerateSucc=lCheckOpen=lOpenLoop= 0;

  lCheckClosed=lClosedLoop=lInsert=
    lInsertLoop=lPush=lPop=lReturnBestNode = 0; 

  NODE *Node, *BestNode; 
  unsigned deltaX, deltaY; 
  OPEN=(NODE *)calloc(1,sizeof( NODE )); 
  if (OPEN == NULL){
    printf("memory allocation error 1\n");
    exit(1);
  }
  CLOSED=(NODE *)calloc(1,sizeof( NODE )); 
  if (CLOSED == NULL){
    printf("memory allocation error 2\n");
    exit(1);
  }

// Setup the Stack. 
  Stack=( STACK *)calloc(1,sizeof(STACK)); 
  if (Stack == NULL){
    printf("memory allocation error\n");
    exit(1);
  }
  Node=(NODE *)calloc(1,sizeof( NODE )); 
  if (Node == NULL){
    printf("memory allocation error\n");
    exit(1);
  }
  Node->g = 0; 
  deltaX = abs(sx-dx);
  deltaY = abs(sy-dy); 
  Node->h = MAX(deltaX, deltaY); 
  Node->f = Node->g + Node->h; 
  Node->x = sx; 
  Node->y = sy; 
  OPEN->NextNode = Node; 
  // make Open List point to first node 
  for (;;) { 
    BestNode = (NODE *)ReturnBestNode(); 
    if (BestNode ==NULL) {
      return NULL;
    }
    // if we've found the end, break and finish 
    if (BestNode->x == dx && BestNode->y == dy){ 
      //printf("Found Goal \n");
      break;
    }
    GenerateSuccessors(BestNode, dx, dy); 
  } 
  return (BestNode); 
} 

NODE* RobotClass::ReturnBestNode(void) { 
  lReturnBestNode += 1; 
  struct NODE *tmp; 
  
  if (OPEN->NextNode == NULL) { 
    printf ("Cannot reach destination\n");
    return NULL; } 
  
  // Pick Node with lowest f, in this case it's the first node in list 
  // because we sort the OPEN list wrt lowest f. Call it BESTNODE. 
  
  tmp = OPEN->NextNode; // point to first node on OPEN 
  OPEN->NextNode = tmp->NextNode; // Make OPEN point to nextnode or NULL. 
  // Next take BESTNODE (or temp in this case) and put it on CLOSED 
  tmp->NextNode = CLOSED->NextNode;
  CLOSED->NextNode = tmp; 
  return(tmp); 
} 

void RobotClass::GenerateSuccessors(NODE *BestNode, int dx, int dy) {
  lGenerateSuccessors += 1; 
  int x, y; 

  // South
  if (Map->FreeTile(x = BestNode->x, y = BestNode->y - 1))
    GenerateSucc(BestNode, x, y, dx, dy); 

  // East 
  if (Map->FreeTile(x = BestNode->x + 1, y = BestNode->y))
    GenerateSucc(BestNode, x, y, dx, dy); 

  // North 
  if (Map->FreeTile(x = BestNode->x , y = BestNode->y + 1))
    GenerateSucc(BestNode, x, y, dx, dy); 

  // West
  if (Map->FreeTile(x = BestNode->x - 1, y = BestNode->y))
    GenerateSucc(BestNode, x, y, dx, dy); 
} 

void RobotClass::GenerateSucc(NODE *BestNode, int x, int y, int dx, int dy) { 

  lGenerateSucc += 1; 
  float g; // total path cost - as stored in the linked lists. 
  int c = 0; 
  unsigned deltaX, deltaY;
  NODE *Old, *Successor; 

  // g(Successor)=g(BestNode)+cost of getting from BestNode to Successor 
  int MoveCost = Map->MapCost(x,y);

  g = BestNode->g + MoveCost; 
  if ((Old = CheckOPEN(x, y)) != NULL){ 
    // ifequal to NULL then not in OPEN list, else it returns the Node in Old 
    for(c = 0; c < 4; c++) 
      if(BestNode->Child[c] == NULL) 
	// Add Old to the list of BestNode's Children (or Successors). 
	break; 
    BestNode->Child[c] = Old; 

    if (g < Old->g) { 
      // if our new g value is < Old's then reset Old's parent to 
      // point to BestNode  
      Old->Parent = BestNode; 
      Old->g = g; 
      Old->f = g + Old->h; 
    } 
  } 
  else if ((Old = CheckCLOSED(x, y)) != NULL){ 
    // if equal to NULL then not in OPEN list, 
    // else it returns the Node in Old 
    for(c = 0; c < 4; c++) 
      if (BestNode->Child[c] == NULL) 
	// Add Old to the list of BestNode's Children (or Successors). 
	break; 
    BestNode->Child[c] = Old; 
    if (g < Old->g) {
      // if our new g value is < Old's 
      // then reset Old's parent to point to BestNode  
      Old->Parent = BestNode; 
      Old->g = g; 
      Old->f = g + Old->h; 
      PropagateDown(Old); 
      // Since we changed the g value of Old, we need 
      // to propagate this new value downwards, i.e. 
      // do a Depth-First traversal of the tree! 
    } 
  } 
  else { 
    Successor = (NODE *) calloc(1,sizeof( NODE )); 
    if (Successor == NULL){
      printf("memory allocation error\n");
      exit(1);
    }
    Successor->Parent = BestNode;
    Successor->g = g; 
    deltaX = abs(x-dx); 
    deltaY = abs(y-dy); 
    Successor->h = MAX(deltaX, deltaY); // directed distance
    Successor->f = g + Successor->h;    // actual cost
    Successor->x = x;
    Successor->y = y; 
    Insert(Successor); // Insert Successor on OPEN list wrt f 
    for(c = 0; c < 4; c++) 
      if (BestNode->Child[c] == NULL) 
	// Add Old to the list of BestNode's Children (or Successors). 
	break;

    BestNode->Child[c] = Successor; 
  } 
} 

NODE* RobotClass::CheckOPEN(int x, int y) { 
  lCheckOpen += 1; 
  NODE *tmp; 
  tmp = OPEN->NextNode; 
  while (tmp != NULL) { 
    lOpenLoop += 1; 
    if (tmp->x == x && tmp->y == y) 
      return (tmp); 
    else 
      tmp = tmp->NextNode; 
  } 
  return (NULL); 
} 

NODE* RobotClass::CheckCLOSED(int x, int y) { 
  lCheckClosed += 1; 
  NODE *tmp;
  tmp = CLOSED->NextNode; 
  while (tmp != NULL) { 
    lClosedLoop += 1; 
    if (tmp->x == x && tmp->y == y) 
      return (tmp); 
    else tmp = tmp->NextNode; 
  }

  return (NULL); 
} 

void RobotClass::Insert(NODE *Successor) { 
  lInsert += 1; 
  NODE *tmp1, *tmp2; 
  float f; 
  if (OPEN->NextNode == NULL) {
    OPEN->NextNode = Successor; 
    return; 
  } 

  // insert into OPEN successor wrt f 
  f = Successor->f; 
  tmp1 = OPEN; 
  tmp2 = OPEN->NextNode; 
  while ((tmp2 != NULL) && (tmp2->f < f)) { 
    lInsertLoop += 1; 
    tmp1 = tmp2; 
    tmp2 = tmp2->NextNode; 
  } 
  Successor->NextNode = tmp2; 
  tmp1->NextNode = Successor; 
} 

void RobotClass::PropagateDown(NODE *Old) { 
  lPropagateDown += 1; 
  int c; 
  float g; 
  NODE *Child, *Father; 
  g = Old->g; 
  int MoveCost;

  // alias. 

  for(c = 0; c < 4; c++) { 
    Child = Old->Child[c]; // Create alias for faster access. 
    if (Child == NULL) 
      break;
    MoveCost = MyMap.MapCost(Child->x, Child->y);
    if (g + MoveCost < Child->g) {
      Child->g = g + MoveCost; 
      Child->f = Child->g + Child->h;
      Child->Parent = Old; 
      // reset parent to new path. 
      Push(Child); 
      // Now the Child's branch need to be 
    } // checked out. Remember the new cost must be propagated down. 
  } 
  while (Stack->NextStackPtr != NULL) {
    lPropLoop += 1; 
    Father = Pop(); 

    for(c = 0; c < 4; c++) { 
      Child = Father->Child[c]; 
      if (Child == NULL) // we may stop the propagation 2 ways: either 
	break;
      MoveCost = MyMap.MapCost(Child->x, Child->y);
      if (Father->g + 1 < Child->g){
	// there are no children, or that the g value of  
	// the child is equal or better than the cost we're propagating 
	Child->g = Father->g + MoveCost; 
	Child->f = Child->g + Child->h; 
	Child->Parent = Father;
	Push(Child); 
      } 
    } 
  } 
}



