#ifndef ROBOT_H
#define ROBOT_H

#include <stdlib.h>
#include <math.h>
#include <stdio.h>

// ASTAR stuff
typedef struct NODE { 
  float f,h; 
  float g,tmpg;
  int x,y; 
  NODE * Parent; 
    NODE *Child[4]; 
  NODE *NextNode;  
}; 

typedef struct STACK { 
  NODE *NodePtr; 
  STACK *NextStackPtr; 
}; 

enum ActionType {Move, Pickup, Drop};

typedef struct Command {
  ActionType Action;
  int Bid;
  int Id;
};

class PackageClass {
 public:
  int ID;
  static PackageClass * FirstPackage;
  PackageClass * NextPackage;
  int Weight;
  int DestX;
  int DestY;
  int XPos;
  int YPos;
  bool DeleteMe;
  bool InTransit;
  float Value;

  PackageClass * AddPackage(int IDp, int Xp, int Yp, 
			    int Dx, int Dy, int W);
  PackageClass * FindPackage(int IDp);
  PackageClass * GetNextPackage(void);
  PackageClass * GetNextPackageTo(int X, int Y);
  void ProcessPackageList(char * List);

};

enum MissionType {MovingToBase = 0, DeliveringPackages = 1, MovingRandom= 3};
enum MoveType {None, Escape, Deathbot, Dumbass};

class RobotClass {
  
public:

  enum DirectionType {North, South, East, West, Stay};

  int GetXPos(void);
  int GetYPos(void);
  inline void SetXPos(int x){XPos = x;};
  inline void SetYPos(int y){YPos = y;};
  void SetMap(MapClass * MapParam);
  void ProcessMovement(DirectionType Direction);
  void DetermineAction(void);
  void Print(void);
  void ProcessPackages(FILE * SockStream);
  void GenerateCommand(FILE * SockStream);
  void SetGoalXY(int XposParam, int YPosParam);
  void CompressPackageList(void);
  void MoveRandom(FILE * SockStream, int Bid=1);
  inline void SetEnergy(int Value) {Energy=Value;};
  inline int GetEnergy(void) {return Energy;};
  inline void SetMaxWeight(int Weight){MaxWeight = Weight;};
  inline int GetCurrentWeight(void) {return CurrentWeight;};
  inline void SetPlayerID(int ID) {PlayerID=ID;};
  inline int GetPlayerID(void) {return PlayerID;};
  PackageClass * Package[20];

  //  AddPackageAtLocation(void);
  //AddPackageCarried(void);
  //DropPackageAtLocation(void);

  MissionType Mission;
  int XPos, YPos;
  int GoalX, GoalY;
  int MaxWeight, CurrentWeight;
  int Energy;
  int PlayerID;
  MapClass * Map;


  NODE *OPEN; 
  NODE *CLOSED; 
  STACK *Stack;

  DirectionType AStarNextMove(int x1, int y1, int x2, int y2, float * cost);
  NODE * FindPath(int sx, int sy, int dx, int dy);
  NODE * ReturnBestNode(void);
  void GenerateSuccessors(NODE *BestNode, int dx, int dy);
  void GenerateSucc(NODE *BestNode, int x, int y, int dx, int dy);
  NODE* RobotClass::CheckOPEN(int x, int y);
  NODE* RobotClass::CheckCLOSED(int x, int y); 
  void Insert(NODE *Successor);
  void PropagateDown(NODE *Old);

  int lPush, lPop, lGenerateSuccessors, lGenerateSucc, lCheckOpen;
  int lOpenLoop, lCheckClosed, lClosedLoop, lInsert, lInsertLoop;
  int lReturnBestNode, lPropLoop, lPropagateDown;

  inline void Push (NODE * Node) {
    lPush += 1;
    STACK * tmp;
    tmp = (STACK *) malloc(sizeof (STACK));
    if (tmp == NULL){
      printf("memory allocation error\n");
      exit(1);
    }
    
    tmp->NodePtr = Node;
    tmp->NextStackPtr = Stack->NextStackPtr;
    Stack->NextStackPtr = tmp;
  };

  inline NODE * Pop(void) {
    lPop += 1;
    NODE * TempNode;
    STACK * TempStack;
    TempStack = Stack->NextStackPtr;
    TempNode = TempStack->NodePtr;
    Stack->NextStackPtr = TempStack->NextStackPtr;
    free(TempStack);
    return(TempNode);
  };

  inline float MAX(unsigned Dx,unsigned Dy) {
    return (float) sqrt((Dx*Dx) + (Dy*Dy));
  };

};



inline int RobotClass::GetXPos(void) {return XPos;};
inline int RobotClass::GetYPos(void) {return YPos;};
inline void RobotClass::SetGoalXY(int XPosParam, int YPosParam) {
  GoalX = XPosParam; GoalY=YPosParam;};
inline void RobotClass::SetMap(MapClass * MapParam) { 
  Map = MapParam;};


class RobotGangClass {
public:
  RobotClass * Robot;
  RobotGangClass * NextRobot;
  static RobotGangClass * FirstRobot;
  bool Updated;
  bool DeleteMe;
  RobotGangClass * RobotAt(int Xp, int Yp);
  static int RobotCount;

  RobotGangClass * FindRobot(int ID);
  RobotGangClass * AddRobot(int ID, MapClass * Map); 
  void CleanList(void);
  void ResetUpdated(void);

  void ProcessStatus(FILE * SockStream);
  void ProcessRobot(char * Line);
};





#endif
